from typing import Union

import jax
import numpy as np
from jax import numpy as jnp

from jaxdf.conv import fd_coefficients_fornberg, reflection_conv
from jaxdf.core import operator
from jaxdf.discretization import Continuous, FiniteDifferences, FourierSeries


def _get_ffts(x):
    if x.real:
        ffts = [jnp.fft.rfft, jnp.fft.irfft]
    else:
        ffts = [jnp.fft.fft, jnp.fft.ifft]
    return ffts


# derivative
@operator
def derivative(x: Continuous, *, axis=0, params=None) -> Continuous:
    r"""Derivative operator for continuous fields.

    Args:
      x: Continuous field
      axis: Axis along which to take the derivative

    Returns:
      Continuous field
    """
    get_x = x.aux["get_field"]

    def grad_fun(p, coords):
        f_jac = jax.jacfwd(get_x, argnums=(1,))
        return jnp.expand_dims(f_jac(p, coords)[0][0][axis], -1)

    return Continuous(x.params, x.domain, grad_fun)


def get_fd_coefficients(
    x: FiniteDifferences, order: int = 1, stagger: Union[float, int] = 0
):
    r"""Returnst the stencil coefficients for a 1D Finite Differences derivative
    operator.

    Args:
      x: FiniteDifferences field
      order: Order of the derivative
      stagger: Stagger of the derivative

    Returns:
      Stencil coefficients
    """

    # Check that all the values of stagger are in [0, 0.5, -0.5]
    # assert stagger in [0, -0.5, 0.5], f'Staggering must be in [0, 0.5, -0.5] for finite differences, got {stagger}'
    accuracy = x.accuracy
    points = np.arange(-accuracy // 2, accuracy // 2 + 1)
    if stagger > 0:
        points = (points + stagger)[:-1]
    elif stagger < 0:
        points = (points + stagger)[1:]

    # get coefficients
    coeffs = fd_coefficients_fornberg(order, points, x0=0)[0].tolist()

    # Append zero if a coefficient has been removed
    if stagger > 0:
        coeffs = coeffs + [0.0]
    elif stagger < 0:
        coeffs = [0.0] + coeffs

    return np.asarray(coeffs)


def fd_derivative_init(x: FiniteDifferences, axis=0, stagger=0, *args, **kwargs):
    r"""Initializes the stencils for FiniteDifferences derivatives. Accepts
    an arbitrary number of positional and keyword arguments after the
    mandatory arguments, which are ignored.

    Args:
      x: FiniteDifferences field
      axis: Axis along which to take the derivative
      stagger: Stagger of the derivative

    Returns:
      Stencil coefficients
    """
    kernel = get_fd_coefficients(x, order=1, stagger=stagger)

    if x.domain.ndim > 1:
        for _ in range(x.domain.ndim - 1):
            kernel = np.expand_dims(kernel, axis=0)
        # Move kernel to the correct axis
        kernel = np.moveaxis(kernel, -1, axis)

    # Add dx
    kernel = kernel / x.domain.dx[axis]

    return kernel


def fd_diag_jacobian_init(x: FiniteDifferences, stagger=[0], *args, **kwargs):
    r"""Initializes the parameters for the diagonal Jacobian of a FiniteDifferences field. Accepts
    an arbitrary number of positional and keyword arguments after the
    mandatory arguments, which are ignored.

    Args:
      x: FiniteDifferences field
      stagger: Stagger of the derivative

    Returns:
      Stencil coefficients

    """
    if len(stagger) != x.domain.ndim:
        stagger = [stagger[0] for _ in range(x.domain.ndim)]

    kernels = []
    for i in range(x.domain.ndim):
        kernels.append(fd_derivative_init(x, axis=i, stagger=stagger[i]))

    return kernels


# gradient
@operator  # type: ignore
def gradient(x: Continuous, *, params=None) -> Continuous:
    r"""Gradient operator for continuous fields.

    Args:
      x: Continuous field

    Returns:
      The gradient of the field
    """
    get_x = x.aux["get_field"]

    def grad_fun(p, coords):
        f_jac = jax.jacfwd(get_x, argnums=(1,))
        v = f_jac(p, coords)[0]
        return v

    return x.update_fun_and_params(x.params, grad_fun)


@operator(init_params=fd_diag_jacobian_init)  # type: ignore
def gradient(x: FiniteDifferences, *, stagger=[0], params=None) -> FiniteDifferences:
    r"""Gradient operator for finite differences fields.

    Args:
      x: FiniteDifferences field
      stagger: Stagger of the derivative

    Returns:
      The gradient of the field
    """
    return diag_jacobian(x, stagger=stagger, params=params)


@operator(init_params=lambda x, *args, **kwargs: {"k_vec": x._freq_axis})  # type: ignore
def gradient(
    x: FourierSeries, *, stagger=[0], correct_nyquist=True, params=None
) -> FourierSeries:
    r"""Gradient operator for Fourier series fields.

    Args:
        x (FourierSeries): Input field
        stagger (list, optional): Staggering value for the returned fields.
          The fields are staggered in the direction of their derivative.
          Defaults to [0].
        correct_nyquist (bool, optional): If `True`, uses a correction of the
          derivative filter for the Nyquist frequency, which preserves Hermitian
          symmetric and null space. See [those notes](https://math.mit.edu/~stevenj/fft-deriv.pdf)
          for more details. Defaults to True.

    Returns:
        FourierSeries: The gradient of the input field.
    """
    assert x.dims == 1  # Gradient only defined for scalar fields

    ffts = _get_ffts(x)
    k_vec = params["k_vec"]

    # Adding staggering
    if len(stagger) == 1 and len(x.domain.N) != 1:
        stagger = stagger * len(x.domain.N)

    dx = x.domain.dx
    k_vec = [
        1j * k * jnp.exp(1j * k * s * delta) for k, delta, s in zip(k_vec, dx, stagger)
    ]

    # Set to zero the filter at the Nyquist frequency
    # if the dimension is even
    # see https://math.mit.edu/~stevenj/fft-deriv.pdf
    if correct_nyquist:
        for f in range(len(k_vec)):
            if x.domain.N[f] % 2 == 0:
                f_nyq = x.domain.N[f] // 2
                k_vec[f] = k_vec[f].at[f_nyq].set(0.0)

    u = x.params[..., 0]

    def single_grad(axis, u):
        u = jnp.moveaxis(u, axis, -1)
        Fx = ffts[0](u, axis=-1)
        iku = Fx * k_vec[axis]
        du = ffts[1](iku, axis=-1, n=u.shape[-1])
        return jnp.moveaxis(du, -1, axis)

    new_params = jnp.stack([single_grad(i, u) for i in range(x.ndim)], axis=-1)
    return FourierSeries(new_params, x.domain)


# diag_jacobian
@operator  # type: ignore
def diag_jacobian(x: Continuous, *, params=None) -> Continuous:
    r"""Diagonal Jacobian operator for continuous fields.

    Args:
      x: Continuous field

    Returns:
      The diagonal Jacobian of the field
    """
    get_x = x.aux["get_field"]

    def diag_fun(p, coords):
        f_jac = jax.jacfwd(get_x, argnums=(1,))
        return jnp.diag(f_jac(p, coords)[0])

    return x.update_fun_and_params(x.params, diag_fun), None


@operator(init_params=fd_diag_jacobian_init)  # type: ignore
def diag_jacobian(
    x: FiniteDifferences, *, stagger=[0], params=None
) -> FiniteDifferences:
    r"""Diagonal Jacobian operator for finite differences fields.

    Args:
      x: FiniteDifferences field
      stagger: Stagger of the derivative

    Returns:
      The diagonal Jacobian of the field
    """
    kernels = params
    array = x.on_grid

    # Apply the corresponding kernel to each dimension
    outs = [reflection_conv(kernels[i], array[..., i]) for i in range(x.ndim)]
    new_params = jnp.stack(outs, axis=-1)

    return x.replace_params(new_params)


@operator(init_params=lambda x, *args, **kwargs: {"k_vec": x._freq_axis})  # type: ignore
def diag_jacobian(
    x: FourierSeries, *, stagger=[0], correct_nyquist=True, params=None
) -> FourierSeries:
    r"""Diagonal Jacobian operator for Fourier series fields.

    Args:
        x (FourierSeries): Input field
        stagger (list, optional): Staggering value for the returned fields.
          The fields are staggered in the direction of their derivative.
          Defaults to [0].
        correct_nyquist (bool, optional): If `True`, uses a correction of the
          derivative filter for the Nyquist frequency, which preserves Hermitian
          symmetric and null space. See [those notes](https://math.mit.edu/~stevenj/fft-deriv.pdf)
          for more details. Defaults to True.

    Returns:
        The vector field whose components are the diagonal entries
          of the Jacobian of the input field.
    """
    ffts = _get_ffts(x)
    k_vec = params["k_vec"]

    # Adding staggering
    if len(stagger) == 1 and len(x.domain.N) != 1:
        stagger = stagger * len(x.domain.N)

    dx = x.domain.dx
    k_vec = [
        1j * k * jnp.exp(1j * k * s * delta) for k, delta, s in zip(k_vec, dx, stagger)
    ]

    # Set to zero the filter at the Nyquist frequency
    # if the dimension is even
    # see https://math.mit.edu/~stevenj/fft-deriv.pdf
    if correct_nyquist:
        for f in range(len(k_vec)):
            if x.domain.N[f] % 2 == 0:
                f_nyq = x.domain.N[f] // 2
                k_vec[f] = k_vec[f].at[f_nyq].set(0.0)

    new_params = jnp.zeros_like(x.params)

    def single_grad(axis, u):
        u = jnp.moveaxis(u, axis, -1)
        Fx = ffts[0](u, axis=-1)
        iku = Fx * k_vec[axis]
        du = ffts[1](iku, axis=-1, n=u.shape[-1])
        return jnp.moveaxis(du, -1, axis)

    for ax in range(x.ndim):
        new_params = new_params.at[..., ax].set(single_grad(ax, x.params[..., ax]))

    return FourierSeries(new_params, x.domain), params


# laplacian
@operator  # type: ignore
def laplacian(x: Continuous, *, params=None) -> Continuous:
    r"""Laplacian operator for continuous fields.

    Args:
      x: Continuous field

    Returns:
      The Laplacian of the field
    """
    get_x = x.aux["get_field"]

    def grad_fun(p, coords):
        hessian = jax.hessian(get_x, argnums=(1,))(p, coords)[0][0][0]
        return jnp.diag(hessian)

    return x.update_fun_and_params(x.params, grad_fun), None


@operator(init_params=lambda x, *args, **kwargs: {"k_vec": x._freq_axis})  # type: ignore
def laplacian(x: FourierSeries, *, params=None):
    r"""Laplacian operator for Fourier series fields.

    Args:
      x (FourierSeries): Input field

    Returns:
      The Laplacian of the field
    """
    assert x.dims == 1  # Laplacian only defined for scalar fields

    ffts = _get_ffts(x)
    k_vec = params["k_vec"]
    u = x.params[..., 0]

    def single_grad(axis, u):
        u = jnp.moveaxis(u, axis, -1)
        Fx = ffts[0](u, axis=-1)
        iku = -Fx * (k_vec[axis] ** 2)
        du = ffts[1](iku, axis=-1, n=u.shape[-1])
        return jnp.moveaxis(du, -1, axis)

    new_params = jnp.sum(
        jnp.stack([single_grad(i, u) for i in range(x.ndim)], axis=-1),
        axis=-1,
        keepdims=True,
    )
    return FourierSeries(new_params, x.domain), params


@operator(init_params=lambda x, *args, **kwargs: {"k_vec": x._freq_axis})  # type: ignore
def heterog_laplacian(
    x: FourierSeries, c: FourierSeries, *, params=None
) -> FourierSeries:
    r"""Computes the position-varying laplacian using algorithm 4 of
    [[Johnson, 2011]](https://math.mit.edu/~stevenj/fft-deriv.pdf).

    Args:
      x (FourierSeries): Input field
      c (FourierSeries): Coefficient field

    Returns:
      The Laplacian of the field
    """
    assert x.dims == 1  # Laplacian only defined for scalar fields

    ffts = _get_ffts(x)
    k_vec = params["k_vec"]
    u = x.params[..., 0]
    v = c.params[..., 0]

    def single_coordinate(axis, u):
        u = jnp.moveaxis(u, axis, -1)
        U = ffts[0](u, axis=-1)
        U_prime = 1j * k_vec[axis] * U

        # Handle Nyquist frequency
        N_on_L = 1 / u.domain.dx[axis]
        N = u.domain.N[axis]
        if N % 2 == 0:
            U_prime = U_prime.at[..., N // 2].set(U[..., N // 2] * N_on_L * jnp.pi * 1j)

        # Multiply with heterogeneous field
        u_prime = ffts[1](U_prime, axis=-1, n=u.shape[-1])
        c_uprime = v * u_prime

        # Get the second derivative
        V = ffts[0](c_uprime, axis=-1)
        V_prime = 1j * k_vec[axis] * V

        # Handle Nyquist frequency
        if N % 2 == 0:
            V_prime = V_prime.at[..., N // 2].set(V[..., N // 2] * N_on_L * jnp.pi * 1j)

        # Return to space
        ddu = ffts[1](V_prime, axis=-1, n=u.shape[-1])

        return jnp.moveaxis(ddu, -1, axis)

    new_params = jnp.sum(
        jnp.stack([single_coordinate(i, u) for i in range(x.ndim)], axis=-1),
        axis=-1,
        keepdims=True,
    )
    return FourierSeries(new_params, x.domain), params


if __name__ == "__main__":
    """# Gets implemented functions
    from jaxdf.util import _get_implemented

    funcs = [
      derivative, diag_jacobian, gradient, laplacian,
    ]

    print('differential.py:')
    print('----------------')
    for f in funcs:
      _get_implemented(f)
    print('\n')
    """



from typing import List

import numpy as np
from jax import numpy as jnp

from jaxdf.conv import reflection_conv
from jaxdf.core import operator
from jaxdf.discretization import (
    Continuous,
    FiniteDifferences,
    FourierSeries,
    OnGrid,
)
from jaxdf.operators.differential import get_fd_coefficients


# compose
@operator  # type: ignore
def compose(x: Continuous, *, params=None):
    r"""Applies function composition on the `get_fun` of the Continuous object."""
    get_x = x.aux["get_field"]

    def decorator(fun):
        def new_fun(p, coord):
            return fun(get_x(p, coord))

        return Continuous(x.params, x.domain, new_fun)

    return decorator


@operator  # type: ignore
def compose(x: OnGrid, *, params=None):
    r"""Maps the given function over the pytree of parameters
    of the `Field`.

    """

    def decorator(fun):
        return x.replace_params(fun(x.params))

    return decorator


@operator  # type: ignore
def compose(x: object, *, params=None):
    r"""For non-field objects, the composition is simply the
    application of the `jax` function to the input.

    ```python
    compose(x)(fun) == fun(x)
    ```
    """

    def decorator(fun):
        return fun(x)

    return decorator


@operator  # type: ignore
def functional(x: object, *, params=None):
    r"""For non-field objects, a functional is simply the
    application of the `jax` function to the input.

    ```python
    functional(x)(fun) == fun(x)
    ```
    """

    def decorator(fun):
        return fun(x)

    return decorator


@operator  # type: ignore
def functional(x: OnGrid, *, params=None):
    r"""Maps the given function over the parameters of the field

    !!! example
        ```python
        x = OnGrid(params=-1.0, ...)
        y = functional(x)(jnp.sum)
        y.params # This is -1.0
        ```
    """

    def decorator(fun):
        return fun(x.params)

    return decorator


# get_component
def get_component(x: OnGrid, *, dim: int, params=None) -> OnGrid:
    r"""Slices the parameters of the field along the last dimensions,
    at the index specified by `dim`.

    Args:
      x: The field to slice
      dim: The index to slice at

    Returns:
      A new 1D field corresponding to the `dim`-th component of the input field.
    """
    new_params = jnp.expand_dims(x.params[..., dim], axis=-1)
    return x.replace_params(new_params)


# shift_operator
@operator  # type: ignore
def shift_operator(x: Continuous, *, dx: object, params=None) -> Continuous:
    r"""Shifts the field by `dx` using function composition.

    Args:
      x: The field to shift
      dx: The shift to apply

    Returns:
      A new field corresponding to the shifted input field.
    """
    get_x = x.aux["get_field"]

    def fun(p, coord):
        return get_x(p, coord + dx)

    return Continuous(x.params, x.domain, fun)


def fd_shift_kernels(x: FiniteDifferences, dx: List[float], *args, **kwargs):
    r"""Computes the shift kernels for FiniteDifferences fields.

    Args:
      x: The field to shift
      dx: The shift to apply

    Returns:
      The kernel to apply to the field coefficients in order to
      shift the field.
    """

    def single_kernel(axis, stagger):
        kernel = get_fd_coefficients(x, order=0, stagger=stagger)
        if x.domain.ndim > 1:
            for _ in range(x.domain.ndim - 1):
                kernel = np.expand_dims(kernel, axis=0)
            # Move kernel to the correct axis
            kernel = np.moveaxis(kernel, -1, axis)
        # kernel = kernel / x.domain.dx[axis]
        return kernel

    stagger = dx[0] / x.domain.dx[0]
    params = []
    for i in range(x.domain.ndim):
        params.append(single_kernel(axis=i, stagger=stagger))

    return params


@operator(init_params=fd_shift_kernels)  # type: ignore
def shift_operator(x: FiniteDifferences, *, dx=[0.0], params=None) -> FiniteDifferences:
    r"""Shifts the field by `dx` using finite differences.

    Args:
      x: The field to shift
      dx: The shift to apply

    Returns:
      A new field corresponding to the shifted input field.
    """
    # Apply convolution
    kernels = params
    array = x.on_grid

    # Apply the corresponding kernel to each dimension
    outs = [reflection_conv(kernels[i], array[..., i], i) for i in range(x.ndim)]
    new_params = jnp.stack(outs, axis=-1)

    return x.replace_params(new_params)


@operator(init_params=lambda x, dx: {"k_vec": x._freq_axis})  # type: ignore
def shift_operator(x: FourierSeries, *, dx=[0], params=None) -> FourierSeries:
    r"""Shifts the field by `dx` using the shift theorem in Fourier space.

    Args:
      x: The field to shift
      dx: The shift to apply

    Returns:
      A new field corresponding to the shifted input field.
    """
    if x.real:
        ffts = [jnp.fft.rfft, jnp.fft.irfft]
    else:
        ffts = [jnp.fft.fft, jnp.fft.ifft]
    k_vec = params["k_vec"]

    # Adding staggering
    if len(dx) == 1 and len(x.domain.N) != 1:
        dx = dx * len(x.domain.N)

    k_vec = [jnp.exp(1j * k * delta) for k, delta in zip(k_vec, dx)]

    new_params = jnp.zeros_like(x.params)

    def single_grad(axis, u):
        u = jnp.moveaxis(u, axis, -1)
        Fx = ffts[0](u, axis=-1)
        iku = Fx * k_vec[axis]
        du = ffts[1](iku, axis=-1, n=u.shape[-1])
        return jnp.moveaxis(du, -1, axis)

    for ax in range(x.ndim):
        new_params = new_params.at[..., ax].set(single_grad(ax, x.params[..., ax]))

    return FourierSeries(new_params, x.domain), params


# sum_over_dims
@operator  # type: ignore
def sum_over_dims(x: Continuous, *, params=None):
    get_x = x.aux["get_field"]

    def fun(p, coords):
        return jnp.sum(get_x(p, coords), axis=-1, keepdims=True)

    return x.update_fun_and_params(x.params, fun)


@operator  # type: ignore
def sum_over_dims(x: OnGrid, *, params=None):
    new_params = jnp.sum(x.params, axis=-1, keepdims=True)
    return x.replace_params(new_params)


if __name__ == "__main__":
    from jaxdf.util import _get_implemented

    print(compose.__name__)
    print(compose.__doc__)
    funcs = [compose, sum_over_dims]

    print("functions.py:")
    print("----------------")
    for f in funcs:
        _get_implemented(f)
    print("\n")



from inspect import getfullargspec
from typing import Callable


class SignatureError(Exception):
    """Exception raised when the signature of a function does not match the
    expected signature.
    """

    pass


def check_fun_has_params(f: Callable):
    r"""Checks that the function `f` expects the argument `params` as
    a mandatory keyword argument

    Args:
      f (Callable): The function to check.

    Raises:
      SignatureError: If the function does not have the argument `params`
    """
    fun_spec = getfullargspec(f)
    if "params" in fun_spec.args:
        raise SignatureError(
            f"The argument 'params' must be a keyword argument in {f}, not a positional argument. Example: def evaluate(x, *, params): ..."
        )
    if "params" not in fun_spec.kwonlyargs:
        raise SignatureError(
            f"The argument 'params' must be a keyword argument in {f}. Example: def evaluate(x, *, params): ..."
        )



from jaxdf.core import operator, params_map
from jaxdf.discretization import Continuous, Field, Linear, OnGrid

"""
This file contains the operators that are
bind with the magic functions of fields
"""


# __add__
@operator  # type: ignore
def __add__(x: Linear, y: Linear, *, params=None):
    new_params = params_map(lambda x, y: x + y, x.params, y.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __add__(x: OnGrid, y: object, *, params=None):
    new_params = params_map(lambda x: x + y, x.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __add__(x: Continuous, y: Continuous, *, params=None):
    get_x = x.aux["get_field"]
    get_y = y.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p[0], coords) + get_y(p[1], coords)

    return Continuous([x.params, y.params], x.domain, get_fun), None


@operator  # type: ignore
def __add__(x: Continuous, y: object, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p, coords) + y

    return Continuous(x.params, x.domain, get_fun), None


# __bool__
@operator  # type: ignore
def __bool__(x: OnGrid, *, params=None):
    new_params = params_map(lambda x: bool(x), x.params)
    return x.replace_params(new_params), None


# __divmod__
@operator  # type: ignore
def __divmod__(x: OnGrid, y: OnGrid, *, params=None):
    new_params = params_map(lambda x, y: divmod(x, y), x.params, y.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __divmod__(x: Linear, y, *, params=None):
    new_params = params_map(lambda x: divmod(x, y), x.params)
    return x.replace_params(new_params), None


# __mul__
@operator  # type: ignore
def __mul__(x: OnGrid, y: OnGrid, *, params=None):
    new_params = params_map(lambda x, y: x * y, x.params, y.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __mul__(x: Linear, y, *, params=None):
    new_params = params_map(lambda x: x * y, x.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __mul__(x: Continuous, y: Continuous, *, params=None):
    get_x = x.aux["get_field"]
    get_y = y.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p[0], coords) * get_y(p[1], coords)

    return x.update_fun_and_params([x.params, y.params], get_fun), None


@operator  # type: ignore
def __mul__(x: Continuous, y, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p, coords) * y

    return x.update_fun_and_params(x.params, get_fun), None


# __neg__
@operator  # type: ignore
def __neg__(x: Linear, *, params=None):
    new_params = params_map(lambda x: -x, x.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __neg__(x: Continuous, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return -get_x(p, coords)

    return Continuous(x.params, x.domain, get_fun), None


# __pow__
@operator  # type: ignore
def __pow__(x: OnGrid, y: OnGrid, *, params=None):
    new_params = params_map(lambda x, y: x**y, x.params, y.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __pow__(x: OnGrid, y: object, *, params=None):
    new_params = params_map(lambda x: x**y, x.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __pow__(x: Continuous, y: Continuous, *, params=None):
    get_x = x.aux["get_field"]
    get_y = y.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p[0], coords) ** get_y(p[1], coords)

    return x.update_fun_and_params([x.params, y.params], get_fun), None


@operator  # type: ignore
def __pow__(x: Continuous, y: object, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p, coords) ** y

    return Continuous(x.params, x.domain, get_fun), None


# __radd__
@operator(precedence=-1)  # type: ignore
def __radd__(x: Field, y: object, *, params=None):
    return x + y, None


# __rmul__
@operator(precedence=-1)  # type: ignore
def __rmul__(x: Field, y: object, *, params=None):
    return x * y, None


# __rpow__
@operator(precedence=-1)  # type: ignore
def __rpow__(x: OnGrid, y: object, *, params=None):
    new_params = params_map(lambda x: y**x, x.params)
    return x.replace_params(new_params), None


# __rsub__
@operator  # type: ignore
def __rsub__(x: Field, y: object, *, params=None):
    return (-x) + y, None


# __rtruediv__
@operator  # type: ignore
def __rtruediv__(x: OnGrid, y: object, *, params=None):
    new_params = params_map(lambda x: y / x, x.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __rtruediv__(x: Continuous, y: object, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return y / get_x(p, coords)

    return Continuous(x.params, x.domain, get_fun), None


# __sub__
@operator  # type: ignore
def __sub__(x: Linear, y: Linear, *, params=None):
    new_params = params_map(lambda x, y: x - y, x.params, y.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __sub__(x: OnGrid, y: object, *, params=None):
    new_params = params_map(lambda x: x - y, x.params)
    return x.replace_params(new_params), None


@operator  # type: ignore
def __sub__(x: Continuous, y: Continuous, *, params=None):
    get_x = x.aux["get_field"]
    get_y = y.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p[0], coords) - get_y(p[1], coords)

    return Continuous([x.params, y.params], x.domain, get_fun), None


@operator  # type: ignore
def __sub__(x: Continuous, y: object, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p, coords) - y

    return Continuous(x.params, x.domain, get_fun), None


# __truediv__
@operator  # type: ignore
def __truediv__(x: Continuous, y: Continuous, *, params=None):
    get_x = x.aux["get_field"]
    get_y = y.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p[0], coords) / get_y(p[1], coords)

    return Continuous([x.params, y.params], x.domain, get_fun), None


@operator  # type: ignore
def __truediv__(x: Continuous, y: object, *, params=None):
    get_x = x.aux["get_field"]

    def get_fun(p, coords):
        return get_x(p, coords) / y

    return Continuous(x.params, x.domain, get_fun), None


@operator  # type: ignore
def __truediv__(x: OnGrid, y: OnGrid, *, params=None):
    new_params = params_map(lambda x, y: x / y, x.params, y.params)
    return x.replace_params(new_params), None


@operator(precedence=-1)  # type: ignore
def __truediv__(x: Linear, y, *, params=None):
    new_params = params_map(lambda x: x / y, x.params)
    return x.replace_params(new_params), None


# inverse
@operator  # type: ignore
def inverse(x: OnGrid, *, params=None):
    new_params = params_map(lambda x: 1 / x, x.params)
    return x.replace_params(new_params), None


if __name__ == "__main__":
    """
    from jaxdf.util import _get_implemented

    magic = [
      __add__, __bool__, __divmod__,
      __mul__, __neg__, __pow__, __radd__,
      __rmul__, __rpow__, __rsub__, __rtruediv__,
      __sub__, __truediv__, inverse
    ]

    print('magic.py:')
    print('----------------')
    for f in magic:
      _get_implemented(f)
    print('\n')
    """



from jax.numpy import expand_dims, ndarray


def append_dimension(x: ndarray):
    return expand_dims(x, -1)


def update_dictionary(old: dict, new_entries: dict):
    r"""Update a dictionary with new entries.

    Args:
      old (dict): The dictionary to update
      new_entries (dict): The new entries to add to the dictionary

    Returns:
      dict: The updated dictionary
    """
    for key, val in zip(new_entries.keys(), new_entries.values()):
        old[key] = val
    return old


def _get_implemented(f):
    r"""Prints the implemented methods of a function. For internal use.

    Arguments:
      f (function): The function to get the implemented methods of.

    Returns:
      None

    """
    from inspect import signature

    # TODO: Why there are more instances for the same types?

    print(f.__name__ + ":")
    instances = []
    a = f.methods.values()
    for f_instance in a:
        instances.append(signature(f_instance[0]).__repr__()[11:-1])

    instances = set(instances)
    for instance in instances:
        # if `self` is in the signature, skip
        if "self" in instance:
            continue
        # Remove `jaxdf.discretization` from instance, wherever
        # it is in the string
        instance = instance.replace("jaxdf.discretization.", "")
        print(" ─ " + instance)



import warnings
from typing import Callable, TypeVar

from jax import eval_shape
from jax import numpy as jnp
from jax import vmap
from jax.tree_util import register_pytree_node_class

from jaxdf.core import Field, new_discretization
from jaxdf.geometry import Domain

PyTree = TypeVar("PyTree")


@new_discretization
class Linear(Field):
    r"""This discretization assumes that the field is a linear function of the
    parameters contained in `Linear.params`.
    """

    def __init__(
        self,
        params: PyTree,
        domain: Domain,
        dims=1,
        aux=None,
    ):
        super().__init__(params, domain, dims, aux)


@register_pytree_node_class
class Continuous(Field):
    r"""A continous discretization, which is defined via a `get_field` function stored
    in the `aux` parameters. Its operations are implemented using function composition
    and autograd.

    !!! example
        ```python
        def get_field(params, x):
          return jnp.tanh(params[0] * x + params[1])

        params = jnp.array([1.0, 2.0])
        domain = Domain((16,), (0.1,))
        a = Continuous(params, domain, get_field)
        ```

    """

    def __init__(self, params: PyTree, domain: Domain, get_fun: Callable):
        r"""Initializes a continuous discretization.

        Args:
          params (PyTree): The parameters of the discretization.
          domain (Domain): The domain of the discretization.
          get_fun (Callable): A function that takes a parameter vector and a point in
          the domain and returns the field at that point. The signature of this
          function is `get_field(params, x)`.

        Returns:
          Continuous: A continuous discretization.
        """
        aux = {"get_field": get_fun}
        x = domain.origin
        dims = eval_shape(get_fun, params, x).shape
        super().__init__(params, domain, dims, aux)

    def tree_flatten(self):
        children = (self.params,)
        aux_data = (self.dims, self.domain, self.aux["get_field"])
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        params = children[0]
        dims, domain, get_fun = aux_data
        a = cls(params, domain=domain, get_fun=get_fun)
        return a

    def replace_params(self, new_params):
        r"""Replaces the parameters of the discretization with new ones. The domain
        and `get_field` function are not changed.

        Args:
          new_params (PyTree): The new parameters of the discretization.

        Returns:
          Continuous: A continuous discretization with the new parameters.
        """
        return self.__class__(new_params, self.domain, self.aux["get_field"])

    def update_fun_and_params(
        self,
        params: PyTree,
        get_field: Callable,
    ):
        r"""Updates the parameters and the function of the discretization.

        Args:
          params (PyTree): The new parameters of the discretization.
          get_field (Callable): A function that takes a parameter vector and a point in
            the domain and returns the field at that point. The signature of this
            function is `get_field(params, x)`.

        Returns:
          Continuous: A continuous discretization with the new parameters and function.
        """
        return self.__class__(params, self.domain, get_field)

    @classmethod
    def from_function(cls, domain, init_fun: Callable, get_field: Callable, seed):
        r"""Creates a continuous discretization from a `get_field` function.

        Args:
          domain (Domain): The domain of the discretization.
          init_fun (Callable): A function that initializes the parameters of the
            discretization. The signature of this function is `init_fun(rng, domain)`.
          get_field (Callable): A function that takes a parameter vector and a point in
            the domain and returns the field at that point. The signature of this
            function is `get_field(params, x)`.
          seed (int): The seed for the random number generator.

        Returns:
          Continuous: A continuous discretization.
        """
        params = init_fun(seed, domain)
        return cls(params, domain=domain, get_fun=get_field)

    def __call__(self, x):
        r"""
        An object of this class can be called as a function, returning the field at a
        desired point.

        !!! example
            ```python
            ...
            a = Continuous.from_function(init_params, domain, get_field)
            field_at_x = a(1.0)
            ```
        """
        return self.get_field(x)

    def get_field(self, x):
        r"""Same as `__call__`."""
        return self.aux["get_field"](self.params, x)

    @property
    def on_grid(self):
        """Returns the field on the grid points of the domain."""
        fun = self.aux["get_field"]
        ndims = len(self.domain.N)
        for _ in range(ndims):
            fun = vmap(fun, in_axes=(None, 0))

        return fun(self.params, self.domain.grid)


@register_pytree_node_class
class OnGrid(Linear):
    r"""A linear discretization on the grid points of the domain."""

    def __init__(
        self,
        params: PyTree,
        domain: Domain,
    ):
        r"""Initializes a linear discretization on the grid points of the domain.

        Args:
          params (PyTree): The parameters of the discretization.
          domain (Domain): The domain of the discretization.

        Returns:
          OnGrid: A linear discretization on the grid points of the domain.
        """
        self.domain = domain
        self.params = params

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, value):
        # Automatically add a dimension for scalar fields, if possible.
        # See this for an explanation of the first if: https://jax.readthedocs.io/en/latest/pytrees.html#custom-pytrees-and-initialization
        if not (type(value) is object or value is None or isinstance(value, OnGrid)):
            if len(value.shape) == len(self.domain.N):
                value = jnp.expand_dims(value, -1)
        self._params = value

    @property
    def dims(self):
        return self.params.shape[-1]

    def tree_flatten(self):
        children = (self.params,)
        aux_data = (self.domain,)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        params = children[0]
        domain = aux_data[0]
        a = cls(params, domain=domain)
        return a

    def __getitem__(self, idx):
        r"""Allow indexing when leading batch / time dimensions are
        present in the parameters

        !!! example
            ```python
            ...
            domain = Domain((16, (1.0,))

            # 10 fields
            params = random.uniform(key, (10, 16, 1))
            a = OnGrid(params, domain)

            # Field at the 5th index
            field = a[5]
            ```

        Returns:
          OnGrid: A linear discretization on the grid points of the domain.

        Raises:
          IndexError: If the field is not indexable (single field).
        """
        if self.ndim + 1 == len(self.params.shape):
            raise IndexError(
                "Indexing is only supported if there's at least one batch / time dimension"
            )

        new_params = self.params[idx]
        return self.__class__(new_params, self.domain)

    def __repr__(self):
        dims = self.dims
        size = self.domain.N
        return f"{self.__class__.__name__}[dims={dims}, size={size}]"

    @classmethod
    def empty(cls, domain, dims=1):
        r"""Creates an empty OnGrid field (zero field). Equivalent to
        `OnGrid(jnp.zeros(domain.N), domain)`.
        """
        N = tuple(
            list(domain.N)
            + [
                dims,
            ]
        )
        return cls(jnp.zeros(N), domain)

    @property
    def is_field_complex(self) -> bool:
        r"""Checks if a field is complex.

        Returns:
          bool: Whether the field is complex.
        """
        return self.params.dtype == jnp.complex64 or self.params.dtype == jnp.complex128

    @property
    def real(self) -> bool:
        r"""Checks if a field is real.

        Returns:
          bool: Whether the field is real.
        """
        return not self.is_field_complex

    @classmethod
    def from_grid(cls, grid_values, domain):
        r"""Creates an OnGrid field from a grid of values.

        Args:
          grid_values (ndarray): The grid of values.
          domain (Domain): The domain of the discretization.
        """
        return cls(grid_values, domain)

    def replace_params(self, new_params):
        r"""Replaces the parameters of the discretization with new ones. The domain
        is not changed.

        Args:
          new_params (PyTree): The new parameters of the discretization.

        Returns:
          OnGrid: A linear discretization with the new parameters.
        """
        return self.__class__(new_params, self.domain)

    @property
    def on_grid(self):
        r"""The field on the grid points of the domain."""
        return self.params


@register_pytree_node_class
class FourierSeries(OnGrid):
    r"""A Fourier series field defined on a collocation grid."""

    def __call__(self, x):
        r"""Uses the Fourier shift theorem to compute the value of the field
        at an arbitrary point. Requires N*2 one dimensional FFTs.

        Args:
          x (float): The point at which to evaluate the field.

        Returns:
          float, jnp.ndarray: The value of the field at the point.
        """
        dx = jnp.asarray(self.domain.dx)
        domain_size = jnp.asarray(self.domain.N) * dx
        shift = x - domain_size / 2 + 0.5 * dx

        k_vec = [jnp.exp(-1j * k * delta) for k, delta in zip(self._freq_axis, shift)]
        ffts = self._ffts

        new_params = self.params

        def single_shift(axis, u):
            u = jnp.moveaxis(u, axis, -1)
            Fx = ffts[0](u, axis=-1)
            iku = Fx * k_vec[axis]
            du = ffts[1](iku, axis=-1, n=u.shape[-1])
            return jnp.moveaxis(du, -1, axis)

        for ax in range(self.ndim):
            new_params = single_shift(ax, new_params)

        origin = tuple([0] * self.ndim)
        return new_params[origin]

    def get_field(self, x):
        warnings.warn(
            "FourierSeries.get_field is deprecated. Use FourierSeries.__call__ instead.",
            DeprecationWarning,
        )
        return self.__call__(x)

    @property
    def _freq_axis(self):
        r"""Returns the frequency axis of the grid"""
        if self.is_field_complex:

            def f(N, dx):
                return jnp.fft.fftfreq(N, dx) * 2 * jnp.pi

        else:

            def f(N, dx):
                return jnp.fft.rfftfreq(N, dx) * 2 * jnp.pi

        k_axis = [f(n, delta) for n, delta in zip(self.domain.N, self.domain.dx)]
        return k_axis

    @property
    def _ffts(self):
        r"""Returns the FFT and iFFT functions that are appropriate for the
        field type (real or complex)
        """
        if self.real:
            return [jnp.fft.rfft, jnp.fft.irfft]
        else:
            return [jnp.fft.fft, jnp.fft.ifft]

    @property
    def _cut_freq_axis(self):
        r"""Same as _freq_axis, but last frequency axis is relative to a real FFT.
        Those frequency axis match with the ones of the rfftn function
        """

        def f(N, dx):
            return jnp.fft.fftfreq(N, dx) * 2 * jnp.pi

        k_axis = [f(n, delta) for n, delta in zip(self.domain.N, self.domain.dx)]
        if not self.is_field_complex:
            k_axis[-1] = (
                jnp.fft.rfftfreq(self.domain.N[-1], self.domain.dx[-1]) * 2 * jnp.pi
            )
        return k_axis

    @property
    def _cut_freq_grid(self):
        return jnp.stack(jnp.meshgrid(*self._cut_freq_axis, indexing="ij"), axis=-1)

    @property
    def _freq_grid(self):
        return jnp.stack(jnp.meshgrid(*self._freq_axis, indexing="ij"), axis=-1)


@register_pytree_node_class
class FiniteDifferences(OnGrid):
    r"""A Finite Differences field defined on a collocation grid."""

    def __init__(
        self,
        params,
        domain,
        accuracy=8,
    ):
        r"""Initializes a Finite Differences field on a collocation grid.
        Args:
          params (PyTree): The parameters of the discretization.
          domain (Domain): The domain of the discretization.
        Returns:
          FiniteDifferences: A Finite Differences field on a collocation grid.
        """
        self.domain = domain
        self.params = params
        self.accuracy = accuracy

    def tree_flatten(self):
        children = (self.params,)
        aux_data = (self.domain, self.accuracy)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        params = children[0]
        domain, accuracy = aux_data
        a = cls(params, domain=domain, accuracy=accuracy)
        return a



from typing import List, Tuple, Union, no_type_check

import numpy as np
from jax import numpy as jnp
from jax import scipy as jsp


def reflection_conv(
    kernel: jnp.ndarray, array: jnp.ndarray, reverse: bool = True
) -> jnp.ndarray:
    r"""Convolves an array with a kernel, using reflection padding.
    The kernel is supposed to have the same number of dimensions as the array.

    Args:
      kernel (jnp.ndarray): The kernel to convolve with.
      array (jnp.ndarray): The array to convolve.
      reverse (bool, optional): Whether to reverse the kernel before convolving.
        Defaults to True.

    Returns:
      The convolved array.
    """
    # Reflection padding the array appropriately
    pad = [((x - 1) // 2, (x - 1) // 2) for x in kernel.shape]
    f = jnp.pad(array, pad, mode="wrap")

    if reverse:
        # Reverse the kernel over all axes
        kernel = jnp.flip(kernel, axis=tuple(range(kernel.ndim)))

    # Apply kernel
    return jsp.signal.convolve(f, kernel, mode="valid")


def bubble_sort_abs_value(
    points_list: List[Union[float, int]]
) -> List[Union[float, int]]:
    r"""Sorts a sequence of grid points by their absolute value.

    Sorting is done __in place__. This function is written with numpy, so it can't
    be transformed by JAX.

    !!! example
        ```python
        lst = [-3, -2, -1, 0, 1, 2, 3]
        sorted_lst = bubble_sort_abs_value(lst)
        print(sorted_lst)
        # [0, 1, -1, 2, -2, 3, -3]
        ```

    Args:
      points_list (List[Union[float, int]]): The grid points to sort.

    Returns:
      The sorted grid points.
    """

    for i in range(len(points_list)):
        for j in range(0, len(points_list) - i - 1):
            magnitude_condition = abs(points_list[j]) > abs(points_list[j + 1])
            same_mag_condition = abs(points_list[j]) == abs(points_list[j + 1])
            sign_condition = np.sign(points_list[j]) < np.sign(points_list[j + 1])
            if magnitude_condition or (same_mag_condition and sign_condition):
                temp = points_list[j]
                points_list[j] = points_list[j + 1]
                points_list[j + 1] = temp

    return points_list


# TODO (astanziola): This fails on mypy for some reason, but can't work out how to fix.
@no_type_check
def fd_coefficients_fornberg(
    order: int, grid_points: List[Union[float, int]], x0: Union[float, int]
) -> Tuple[List[None], List[Union[float, int]]]:
    r"""Generate finite difference stencils for a given order and grid points, using
    the Fornberg algorithm described in [[Fornberg, 2018]](https://web.njit.edu/~jiang/math712/fornberg.pdf).

    The grid points can be placed in any order, can be at arbitrary locations (for example, to implemente staggered
    stencils) and don't need to be equidistant.
    The stencil is evaluated for a point in `x0`. Note that setting `order=0` will generate interpolation coefficients
    for the point `x0`.

    !!! example
        ```python
        grid_points = [0, 1, 2, -1, -2]
        x0 = 0.0
        order = 2
        stencil, nodes = fd_coefficients_fornberg(order, grid_points, x0)
        print(f"Stencil: {stencil}, Nodes: {nodes}")
        # Stencil: [-0.08333333  1.33333333 -2.5         1.33333333 -0.08333333], Nodes: [-2 -1  0  1  2]
        ```

    Args:
      order (int): The order of the stencil.
      grid_points (List[Union[float, int]]): The grid points to use.
      x0 (Union[float, int]): The point at which to evaluate the stencil.

    Returns:
      The stencil and the grid points where the stencil is evaluated.
    """
    # from Generation of Finite Difference Formulas on Arbitrarily Spaced Grids
    # Bengt Fornberg, 1998
    # https://web.njit.edu/~jiang/math712/fornberg.pdf
    M = order
    N = len(grid_points) - 1

    # Sort the grid points
    alpha = bubble_sort_abs_value(grid_points)
    delta = dict()  # key: (m,n,v)
    delta[(0, 0, 0)] = 1.0
    c1 = 1.0

    for n in range(1, N + 1):
        c2 = 1.0
        for v in range(n):
            c3 = alpha[n] - alpha[v]
            c2 = c2 * c3
            if n < M:
                delta[(n, n - 1, v)] = 0.0
            for m in range(min([n, M]) + 1):
                d1 = delta[(m, n - 1, v)] if (m, n - 1, v) in delta.keys() else 0.0
                d2 = (
                    delta[(m - 1, n - 1, v)]
                    if (m - 1, n - 1, v) in delta.keys()
                    else 0.0
                )
                delta[(m, n, v)] = ((alpha[n] - x0) * d1 - m * d2) / c3

        for m in range(min([n, M]) + 1):
            d1 = (
                delta[(m - 1, n - 1, n - 1)]
                if (m - 1, n - 1, n - 1) in delta.keys()
                else 0.0
            )
            d2 = delta[(m, n - 1, n - 1)] if (m, n - 1, n - 1) in delta.keys() else 0.0
            delta[(m, n, n)] = (c1 / c2) * (m * d1 - (alpha[n - 1] - x0) * d2)
        c1 = c2

    # Extract the delta with m = M and n = N
    coeffs = [None] * (N + 1)
    for key in delta:
        if key[0] == M and key[1] == N:
            coeffs[key[2]] = delta[key]

    # sort coefficeient and alpha by alpha
    idx = np.argsort(alpha)
    alpha = np.take_along_axis(np.asarray(alpha), idx, axis=-1)
    coeffs = np.take_along_axis(np.asarray(coeffs), idx, axis=-1)

    return coeffs, alpha



# nopycln: file
from jaxdf.core import operator, debug_config, constants  # isort:skip
from jaxdf import util, geometry  # isort:skip
from jaxdf.discretization import *  # isort:skip

# Must be imported after discretization
from jaxdf.operators.magic import *  # isort:skip
from jaxdf import operators  # isort:skip



from jax import numpy as jnp

from jaxdf.core import operator
from jaxdf.discretization import OnGrid

from .functions import compose


@operator
def dot_product(x: OnGrid, y: OnGrid, *, params=None):
    r"""Computes the dot product of two fields."""
    x_conj = compose(x)(jnp.conj)
    return jnp.sum((x_conj * y).on_grid)



"""
This module contains dummy operators for testing purposes.
"""
from jaxdf.core import operator
from jaxdf.discretization import Continuous, Field, OnGrid


@operator  # type: ignore
def dummy(x: Field, *, params=None):
    """Dummy operator for testing purposes."""
    return x


@operator  # type: ignore
def dummy(x: OnGrid, *, params=None):
    r"""A dummy operator that is useful for debugging."""
    if params is None:
        params = {"k": 3.0}
    return params["k"] * x


@operator  # type: ignore
def dummy(x: Continuous, *, params=None):
    if params is None:
        params = {"k": 3.0}
    get_x = x.aux["get_field"]

    def get_fun(p__par, coords):
        p, params = p__par
        return get_x(p, coords) + params["k"]

    return x.update_fun_and_params([x.params, params], get_fun), params


def yummy_init(x: OnGrid, *args, **kwargs):
    return {"k": 3.0}


@operator(init_params=yummy_init)  # type: ignore
def yummy(x: OnGrid, *, params=None):
    r"""A dummy operator that is useful for debugging."""
    return params["k"] * x



import types
from functools import wraps
from typing import Callable, Union

from jax.tree_util import register_pytree_node_class, tree_map
from plum import Dispatcher, type_of

from jaxdf.exceptions import check_fun_has_params

# Initialize the dispatch table
_jaxdf_dispatch = Dispatcher()

# Configuration
debug_config = {
    "debug_dispatch": False,
}


def _operator(evaluate, precedence, init_params):
    check_fun_has_params(evaluate)

    # If the parameter initialization function is not provided, then
    # assume that the operator has no parameters
    if init_params is None:

        def init_params(*args, **kwargs):
            return None

    # Create the operator function
    @wraps(evaluate)
    def wrapper(*args, **kwargs):
        # Check if the parameters are not passed
        if "params" not in kwargs:
            kwargs["params"] = init_params(*args, **kwargs)

        if debug_config["debug_dispatch"]:
            print(
                f"Dispatching {evaluate.__name__} with for types {evaluate.__annotations__}"
            )

        outs = evaluate(*args, **kwargs)
        if isinstance(outs, tuple) and len(outs) > 1:
            # Overload the field class with an extra attribute
            field = outs[0]
        else:
            field = outs

        return field

    # Adds the parameters initializer to the functin wrapper
    wrapper._initialize_parameters = init_params

    # Register the operator in the dispatch table
    f = _jaxdf_dispatch(wrapper, precedence=precedence)

    # Bind an default_params method that returns the default parameters
    def _bound_init_params(self, *args, **kwargs):
        # the method is resolved only on non-keyword arguments,
        # see: https://github.com/wesselb/plum/issues/40#issuecomment-1321164488
        self._resolve_pending_registrations()
        sig_types = tuple([type_of(x) for x in args])

        # """ # TODO: This is a hack that shouldn't be needed
        # if not self._runtime_type_of:
        #    sig_types = tuple([type(x) for x in args])
        # else:
        #    sig_types = tuple([promised_type_of.resolve()(x) for x in args])
        # """

        method, _ = self.resolve_method(*sig_types)
        return method._initialize_parameters(*args, **kwargs)

    f.default_params = types.MethodType(_bound_init_params, f)

    return f


def operator(
    evaluate: Union[Callable, None] = None,
    init_params: Union[Callable, None] = None,
    precedence: int = 0,
):
    r"""Decorator for defining operators using multiple dispatch. The type annotation of the
    `evaluate` function are used to determine the dispatch rules. The dispatch syntax is the
    same as the Julia one, that is: operators are dispatched on the types of the positional arguments.
    Keyword arguments are not considered for dispatching.

    Keyword arguments are defined after the `*` in the function signature.

    !!! example
        ```python
        @operator
        def my_operator(x: FourierSeries, *, dx: float, params=None):
          ...
        ```

    The argument `params` is mandatory and it must be a keyword argument. It is used to pass the
    parameters of the operator, for example the stencil coefficients of a finite difference operator.

    The default value of the parameters is specified by the `init_params` function, as follows:

    !!! example
        ```python

        def params_initializer(x, *, dx):
          return {"stencil": jnp.ones(x.shape) * dx}

        @operator(init_params=params_initializer)
        def my_operator(x, *, dx, params=None):
          b = params["stencil"] / dx
          y_params = jnp.convolve(x.params, b, mode="same")
          return x.replace_params(y_params)
        ```

    The default value of `params` is not considered during computation.
    If the operator has no parameters, the `init_params` function can be omitted. In this case, the
    `params` value is set to `None`.

    For constant parameters, the `constants` function can be used:

    !!! example
        ```python
        @operator(init_params=constants({"a": 1, "b": 2.0}))
        def my_operator(x, *, params):
          return x + params["a"] + params["b"]
        ```


    Args:
      evaluate (Callable): A function with the signature `evaluate(field, *args, **kwargs, params)`.
        It must return a tuple, with the first element being a field and the second
        element being the default parameters for the operator.
      init_params (Callable): A function that overrides the default parameters initializer for the
        operator. Useful when running the operator just to get the parameters is expensive.
      precedence (int): The precedence of the operator if an ambiguous match is found.

    Returns:
      Callable: The operator function with signature `evaluate(field, *args, **kwargs, params)`.

    """
    if evaluate is None:
        # Returns the decorator
        def decorator(evaluate):
            return _operator(evaluate, precedence, init_params)

        return decorator
    else:
        return _operator(evaluate, precedence, init_params)


def new_discretization(cls):
    r"""Wrapper around `jax.tree_util.register_pytree_node_class` that can
    be used to register a new discretization.

    If the discretization doesn't have the same `__init__` function as the
    parent class, the methods `tree_flatten` and `tree_unflatten` must be
    present (see [Extending pytrees](https://jax.readthedocs.io/en/latest/pytrees.html)
    in the JAX documentation).

    !!! example
        ```python
        @new_discretization
        class Polynomial(Continuous):
          @classmethod
          def from_params(cls, params, domain):
            def get_fun(params, x):
              i = jnp.arange(len(params))
              powers = x**i
              return jnp.expand_dims(jnp.sum(params*(x**i)), -1)
            return cls(params, domain, get_fun)
        ```
    """
    return register_pytree_node_class(cls)


def constants(value) -> Callable:
    r"""This is a higher order function for defining constant parameters of
    operators, independent of the operator arguments.

    !!! example

        ```python
        @operator(init_params=constants({"a": 1, "b": 2.0}))
        def my_operator(x, *, params):
          return x + params["a"] + params["b"]
        ```

    Args:
      value (Any): The value of the constant parameters.

    Returns:
      Callable: The parameters initializer function that returns the constant value.
    """

    def init_params(*args, **kwargs):
        return value

    return init_params


@new_discretization
class Field(object):
    r"""The base-class for all discretizations. This class is also responsible
    for binding the operators in `jaxdf.operators.magic` to
    the magic methods of the discretization.

    Normally you should not use this class directly, but instead use the
    `new_discretization` decorator to register
    a new discretization based on this class.
    """

    def __init__(
        self,
        params,
        domain,
        dims=1,
        aux=None,
    ):
        self.params = params
        self.domain = domain
        self.dims = dims
        self.aux = aux

    def tree_flatten(self):
        children = (self.params,)
        aux_data = (self.dims, self.domain, self.aux)
        return (children, aux_data)

    @property
    def ndim(self):
        return self.domain.ndim

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        params = children[0]
        dims, domain, aux = aux_data
        a = cls(params, dims=dims, domain=domain, aux=aux)
        return a

    def __repr__(self):  #
        classname = self.__class__.__name__
        return f"Field {classname}"

    def __str__(self):
        return self.__repr__()

    def replace_params(self, new_params):
        r"""Returns a new field of the same type, with the same domain and auxiliary data, but with new parameters.

        !!! example
            ```python
            x = FourierSeries(jnp.ones(10), domain=domain)
            y_params = x.params + 1
            y = x.replace_params(y_params)
            ```

        Args:
          new_params (Any): The new parameters.

        Returns:
          Field: A new field with the same domain and auxiliary data, but with new parameters.
        """
        return self.__class__(new_params, self.domain, self.dims, self.aux)

    # Dummy magic functions to make it work with
    # the dispatch system
    def __add__(self, other):
        return __add__(self, other)

    def __radd__(self, other):
        return __radd__(self, other)

    def __sub__(self, other):
        return __sub__(self, other)

    def __rsub__(self, other):
        return __rsub__(self, other)

    def __mul__(self, other):
        return __mul__(self, other)

    def __rmul__(self, other):
        return __rmul__(self, other)

    def __neg__(self):
        return __neg__(self)

    def __pow__(self, other):
        return __pow__(self, other)

    def __rpow__(self, other):
        return __rpow__(self, other)

    def __truediv__(self, other):
        return __truediv__(self, other)

    def __rtruediv__(self, other):
        return __rtruediv__(self, other)


@operator
def __add__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __radd__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __sub__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __rsub__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __mul__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __rmul__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __neg__(self, *, params=None):
    raise NotImplementedError(f"Function not implemented for {type(self)}")


@operator
def __pow__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __rpow__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __truediv__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )


@operator
def __rtruediv__(self, other, *, params=None):
    raise NotImplementedError(
        f"Function not implemented for {type(self)} and {type(other)}"
    )

    # Lifted jax functions for convenience


def params_map(f: Callable, field: Field, *rest) -> Field:
    r"""Maps a function to the parameters of a Field.

    Since a Field is a pytree, this is equivalent to (and implemented
    using) `jax.tree_util.tree_map`

    Returns a field with the same type of `f`, with updated
    parameters

    !!! danger
        This function has been deprecated. Use `jaxdf.operators.functions.compose` instead

    Args:
      f (Callable): A function that is applied to all the leaves of the
        parameters of the input fields
      field (Field): The field to map the function to
      *rest: Optional additional fields to map the function to

    Returns:
      Field: A field with the same discretization as `field`, with updated parameters.
    """
    return tree_map(f, field, *rest)



from functools import reduce
from typing import Iterable, NamedTuple

import jax
from jax import numpy as jnp
from jax import random


class Domain(NamedTuple):
    r"""Domain class describing a rectangular domain

    Attributes:
        size (Tuple[int]): The size of the domain in absolute units.
        dx (Tuple(float)): The unit of measure
    """
    N: Iterable[int] = (32, 32)
    dx: Iterable[float] = (1.0, 1.0)

    @property
    def size(self):
        r"""The lenght of the grid sides

        Returns:
            Tuple[float]: The size of the domain, in absolute units

        """
        return list(map(lambda x: x[0] * x[1], zip(self.N, self.dx)))

    @property
    def ndim(self):
        r"""The number of dimensions of the domain

        Returns:
            int: The number of dimensions of the domain

        """
        return len(self.N)

    @property
    def cell_volume(self):
        r"""The volume of a single cell

        Returns:
            float: The volume of a single cell

        """
        return reduce(lambda x, y: x * y, self.dx)

    @property
    def spatial_axis(self):
        r"""The spatial axis of the domain

        Returns:
            Tuple[jnp.array]: The spatial axis of the domain

        """

        def _make_axis(n, delta):
            if n % 2 == 0:
                return jnp.arange(0, n) * delta - delta * n / 2
            else:
                return jnp.arange(0, n) * delta - delta * (n - 1) / 2

        axis = [_make_axis(n, delta) for n, delta in zip(self.N, self.dx)]
        axis = [ax - jnp.mean(ax) for ax in axis]
        return axis

    @property
    def boundary_sampler(self):
        r"""Returns a function that samples a point on the boundary of the domain

        Returns:
            Callable: A function that samples a point on the boundary of the domain.
                This function takes a seed and an integer number of samples and returns
                a list of samples.

        !!! example
            ```python
            >>> domain = Domain((10, 10), (0.1, 0.1))
            >>> boundary_sampler = domain.boundary_sampler
            >>> boundary_sampler(random.PRNGKey(0), 10)
            Array([[-0.02072918,  0.5       ],
                   [-0.5       ,  0.49063694],
                   [-0.18872023, -0.5       ],
                   [ 0.31801188, -0.5       ],
                   [-0.1319474 , -0.5       ],
                   [ 0.5       , -0.36944878],
                   [ 0.5       ,  0.46956718],
                   [ 0.4608934 , -0.5       ],
                   [-0.09031796,  0.5       ],
                   [-0.5       ,  0.40659428]], dtype=float32)

            ```

        """
        L = jnp.asarray(self.size) / 2

        def sample(seed):
            seeds = random.split(seed, 3)
            first = 2 * jnp.expand_dims(random.uniform(seeds[0]), 0) - 1
            others = 2 * random.bernoulli(seeds[1], shape=(self.ndim - 1,)) - 1
            sample = jnp.concatenate([first, others]).astype(jnp.float32)
            random_perm = random.permutation(seeds[2], sample)
            sample = random_perm * L
            return sample

        def multi_samples(seed, num_samples: int):
            seeds = random.split(seed, num_samples)
            return jax.vmap(sample)(seeds)

        return multi_samples

    @property
    def domain_sampler(self):
        r"""Returns a function that samples a point in the domain

        Returns:
            Callable: A function that samples a point in the domain.
                This function takes a seed and an integer number of samples and returns
                a list of samples.

        !!! example
            ```python
            >>> domain = Domain((10, 10), (0.1, 0.1))
            >>> domain_sampler = domain.domain_sampler
            >>> domain_sampler(random.PRNGKey(0), 10)
            Array([[ 0.06298566,  0.35970068],
                   [-0.20049798,  0.05455852],
                   [ 0.33402848, -0.04824698],
                   [ 0.27945423,  0.2805649 ],
                   [ 0.49464726,  0.3473643 ],
                   [-0.16299951, -0.27665186],
                   [-0.06442916,  0.04995835],
                   [ 0.05011427, -0.17267668],
                   [-0.39805043, -0.05386746],
                   [ 0.46900105,  0.21520817]], dtype=float32)

            ```

        """
        L = jnp.asarray(self.size) / 2

        def sample(seed):
            sample = 2 * random.uniform(seed, shape=(self.ndim,)) - 1
            return sample * L

        def multi_samples(seed, num_samples: int):
            seeds = random.split(seed, num_samples)
            return jax.vmap(sample)(seeds)

        return multi_samples

    @property
    def origin(self):
        return jnp.zeros((self.ndim,))

    @staticmethod
    def _make_grid_from_axis(axis):
        return jnp.stack(jnp.meshgrid(*axis, indexing="ij"), axis=-1)

    @property
    def grid(self):
        """A grid of spatial position, of size
        `Nx x Ny x Nz x ... x num_axis` such that the element
        `[x1,x2,x3, .., :]` is a coordinate vector.

        Returns:
            jnp.array: A grid of spatial position

        """
        axis = self.spatial_axis
        return self._make_grid_from_axis(axis)
